from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.classes import Link

from .icons import (
    icon_queue_task_type_list, icon_worker_list, icon_worker_queue_list
)
from .permissions import permission_task_view

link_queue_task_type_list = Link(
    icon=icon_queue_task_type_list, kwargs={
        'queue_name': 'resolved_object.name'
    }, permission=permission_task_view, text=_(message='Task type list'),
    view='task_manager:queue_task_type_list'
)
link_worker_list = Link(
    icon=icon_worker_list, permission=permission_task_view,
    text=_(message='Worker list'), view='task_manager:worker_list'
)
link_worker_queue_list = Link(
    icon=icon_worker_queue_list, kwargs={
        'worker_name': 'resolved_object.name'
    }, permission=permission_task_view, text=_(message='Queue list'),
    view='task_manager:worker_queue_list'
)



from mayan.apps.documents.tests.base import GenericDocumentTestCase

from ..models import DuplicateBackendEntry

from .mixins import DuplicatedDocumentTaskTestMixin


class DuplicatedDocumentTaskTestCase(
    DuplicatedDocumentTaskTestMixin, GenericDocumentTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._upload_test_document()
        self._upload_test_document()
        DuplicateBackendEntry.objects.all().delete()

    def test_task_duplicates_scan_all(self):
        self._clear_events()

        self._execute_task_duplicates_scan_all()

        self.assertEqual(
            list(
                DuplicateBackendEntry.objects.get_duplicates_of(
                    document=self._test_documents[0]
                )
            ), [self._test_documents[1]]
        )
        self.assertEqual(
            list(
                DuplicateBackendEntry.objects.get_duplicates_of(
                    document=self._test_documents[1]
                )
            ), [self._test_documents[0]]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_task_duplicates_scan_for(self):
        self._clear_events()

        self._execute_task_duplicates_scan_for()

        self.assertEqual(
            list(
                DuplicateBackendEntry.objects.get_duplicates_of(
                    document=self._test_documents[0]
                )
            ), [self._test_documents[1]]
        )
        self.assertEqual(
            list(
                DuplicateBackendEntry.objects.get_duplicates_of(
                    document=self._test_documents[1]
                )
            ), [self._test_documents[0]]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from mayan.apps.documents import storages
from mayan.apps.storage.tests.mixins import StorageSettingTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from ..literals import (
    STORAGE_NAME_DOCUMENT_FILE_PAGE_IMAGE_CACHE, STORAGE_NAME_DOCUMENT_FILES,
    STORAGE_NAME_DOCUMENT_VERSION_PAGE_IMAGE_CACHE
)
from ..settings import (
    setting_document_file_page_image_cache_maximum_size,
    setting_document_file_page_image_cache_storage_backend_arguments,
    setting_document_file_storage_backend_arguments,
    setting_document_version_page_image_cache_maximum_size,
    setting_document_version_page_image_cache_storage_backend_arguments,
    setting_language_codes
)


class DocumentSettingsTestCase(BaseTestCase):
    def test_documents_language_codes_setting_double_quotes(self):
        self._set_environment_variable(
            name='MAYAN_{}'.format(setting_language_codes.global_name),
            value='["spa","fra"]'
        )

        self.assertEqual(
            setting_language_codes.value,
            ['spa', 'fra']
        )

    def test_documents_language_codes_setting_single_quotes(self):
        self._set_environment_variable(
            name='MAYAN_{}'.format(setting_language_codes.global_name),
            value="['spa','deu']"
        )

        self.assertEqual(
            setting_language_codes.value,
            ['spa', 'deu']
        )


class DocumentFileStorageSettingsTestCase(
    StorageSettingTestMixin, BaseTestCase
):
    def test_setting_document_file_storage_backend_arguments_invalid_value(self):
        assertion = self._test_storage_setting_with_invalid_value(
            setting=setting_document_file_storage_backend_arguments,
            storage_module=storages,
            storage_name=STORAGE_NAME_DOCUMENT_FILES
        )

        self.assertTrue(
            'Unable to initialize' in str(assertion.exception)
        )
        self.assertTrue(
            'document file storage' in str(assertion.exception)
        )

    def test_setting_document_file_page_image_cache_maximum_size(self):
        old_value = setting_document_file_page_image_cache_maximum_size.value
        new_value = old_value + 1
        setting_document_file_page_image_cache_maximum_size.value = '{}'.format(new_value)

        self.assertEqual(
            setting_document_file_page_image_cache_maximum_size.value,
            new_value
        )

    def test_setting_document_file_page_image_cache_storage_backend_arguments_invalid_value(self):
        assertion = self._test_storage_setting_with_invalid_value(
            setting=setting_document_file_page_image_cache_storage_backend_arguments,
            storage_module=storages,
            storage_name=STORAGE_NAME_DOCUMENT_FILE_PAGE_IMAGE_CACHE
        )

        self.assertTrue(
            'Unable to initialize' in str(assertion.exception)
        )
        self.assertTrue(
            'document file image storage' in str(assertion.exception)
        )


class DocumentVersionStorageSettingsTestCase(
    StorageSettingTestMixin, BaseTestCase
):
    def test_setting_document_version_page_image_cache_storage_backend_arguments_invalid_value(self):
        assertion = self._test_storage_setting_with_invalid_value(
            setting=setting_document_version_page_image_cache_storage_backend_arguments,
            storage_module=storages,
            storage_name=STORAGE_NAME_DOCUMENT_VERSION_PAGE_IMAGE_CACHE
        )

        self.assertTrue(
            'Unable to initialize' in str(assertion.exception)
        )
        self.assertTrue(
            'document version image storage' in str(assertion.exception)
        )

    def test_setting_document_version_page_image_cache_maximum_size_callback(self):
        old_value = setting_document_version_page_image_cache_maximum_size.value
        new_value = old_value + 1
        setting_document_version_page_image_cache_maximum_size.value = '{}'.format(new_value)

        self.assertEqual(
            setting_document_version_page_image_cache_maximum_size.value,
            new_value
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Dependencies'), name='dependencies'
)

permission_dependencies_view = namespace.add_permission(
    label=_(message='View dependencies'), name='dependencies_view'
)



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('events', '0006_objecteventsubscription')
    ]

    operations = [
        migrations.AlterModelOptions(
            name='notification',
            options={
                'ordering': ('-action__timestamp',),
                'verbose_name': 'Notification',
                'verbose_name_plural': 'Notifications'
            }
        )
    ]



from mayan.apps.dependencies.classes import PythonDependency

PythonDependency(
    module=__name__, name='boto3', version_string='==1.33.7'
)
PythonDependency(
    module=__name__, name='django-storages', version_string='==1.14.2'
)
PythonDependency(
    module=__name__, name='extract-msg', version_string='==0.48.0'
)
PythonDependency(
    module=__name__, name='pycryptodome', version_string='==3.20.0'
)



class FormSetExtraFormKwargsMixin:
    def __init__(self, *args, **kwargs):
        self.form_extra_kwargs = kwargs.pop(
            'form_extra_kwargs', {}
        )
        super().__init__(*args, **kwargs)

    def get_form_kwargs(self, index):
        form_kwargs = super().get_form_kwargs(index=index)
        form_kwargs.update(self.form_extra_kwargs)
        return form_kwargs



from mayan.apps.source_web_forms.tests.mixins import WebFormSourceTestMixin
from mayan.apps.sources.tests.mixins.source_view_mixins import SourceActionViewTestMixin
from mayan.apps.testing.tests.mixins import TestMixinObjectCreationTrack

from ..models import Tag

from .literals import (
    TEST_TAG_COLOR, TEST_TAG_COLOR_EDITED, TEST_TAG_LABEL,
    TEST_TAG_LABEL_EDITED
)


class TagTestMixin(TestMixinObjectCreationTrack):
    _test_object_model = Tag
    _test_object_name = '_test_tag'
    _test_tag_add_test_document = False
    auto_create_test_tag = False

    def setUp(self):
        super().setUp()
        self._test_tag_list = []

        if self.auto_create_test_tag:
            self._create_test_tag(
                add_test_document=self._test_tag_add_test_document
            )

    def _create_test_tag(self, add_test_document=False):
        total_test_labels = len(self._test_tag_list)
        label = '{}_{}'.format(TEST_TAG_LABEL, total_test_labels)

        self._test_tag = Tag.objects.create(
            color=TEST_TAG_COLOR, label=label
        )

        self._test_tag_list.append(self._test_tag)

        if add_test_document:
            self._test_tag.attach_to(
                document=self._test_document, user=self._test_case_user
            )


class DocumentTagViewTestMixin(TagTestMixin):
    def _request_test_document_tag_attach_view(self):
        return self.post(
            viewname='tags:tag_attach', kwargs={
                'document_id': self._test_document.pk
            }, data={
                'tags': self._test_tag.pk,
                'user': self._test_case_user.pk
            }
        )

    def _request_test_document_multiple_tag_multiple_attach_view(self):
        return self.post(
            viewname='tags:multiple_documents_tag_attach', data={
                'id_list': self._test_document.pk, 'tags': self._test_tag.pk,
                'user': self._test_case_user.pk
            }
        )

    def _request_test_document_tag_multiple_remove_view(self):
        return self.post(
            viewname='tags:single_document_multiple_tag_remove', kwargs={
                'document_id': self._test_document.pk
            }, data={
                'tags': self._test_tag.pk
            }
        )

    def _request_test_document_multiple_tag_remove_view(self):
        return self.post(
            viewname='tags:multiple_documents_selection_tag_remove', data={
                'id_list': self._test_document.pk,
                'tags': self._test_tag.pk
            }
        )

    def _request_test_document_tag_list_view(self):
        return self.get(
            viewname='tags:document_tag_list', kwargs={
                'document_id': self._test_document.pk
            }
        )

    def _request_test_tag_document_list_view(self):
        return self.get(
            viewname='tags:tag_document_list', kwargs={
                'tag_id': self._test_tag.pk
            }
        )


class TagAPIViewTestMixin(TagTestMixin):
    def _request_test_document_tag_attach_api_view(self):
        return self.post(
            viewname='rest_api:document-tag-attach', kwargs={
                'document_id': self._test_document.pk
            }, data={'tag': self._test_tag.pk}
        )

    def _request_test_document_tag_list_api_view(self):
        return self.get(
            viewname='rest_api:document-tag-list', kwargs={
                'document_id': self._test_document.pk
            }
        )

    def _request_test_document_tag_remove_api_view(self):
        return self.post(
            viewname='rest_api:document-tag-remove', kwargs={
                'document_id': self._test_document.pk
            }, data={'tag': self._test_tag.pk}
        )

    def _request_test_tag_create_api_view(self):
        self._test_object_track()

        response = self.post(
            viewname='rest_api:tag-list', data={
                'label': TEST_TAG_LABEL, 'color': TEST_TAG_COLOR
            }
        )

        self._test_object_set()

        return response

    def _request_test_tag_delete_api_view(self):
        return self.delete(
            viewname='rest_api:tag-detail',
            kwargs={'tag_id': self._test_tag.pk}
        )

    def _request_test_tag_detail_api_view(self):
        return self.get(
            viewname='rest_api:tag-detail',
            kwargs={'tag_id': self._test_tag.pk}
        )

    def _request_test_tag_document_list_api_view(self):
        return self.get(
            viewname='rest_api:tag-document-list', kwargs={
                'tag_id': self._test_tag.pk
            }
        )

    def _request_test_tag_edit_api_view(self, extra_data=None, verb='patch'):
        data = {
            'color': TEST_TAG_COLOR_EDITED,
            'label': TEST_TAG_LABEL_EDITED
        }

        if extra_data:
            data.update(extra_data)

        return getattr(self, verb)(
            viewname='rest_api:tag-detail',
            kwargs={'tag_id': self._test_tag.pk},
            data=data
        )

    def _request_test_tag_list_api_view(self):
        return self.get(viewname='rest_api:tag-list')


class TagViewTestMixin(TagTestMixin):
    def _request_test_tag_create_view(self):
        self._test_object_track()

        response = self.post(
            viewname='tags:tag_create', data={
                'label': TEST_TAG_LABEL,
                'color': TEST_TAG_COLOR
            }
        )

        self._test_object_set()

        return response

    def _request_test_tag_delete_view(self):
        return self.post(
            viewname='tags:tag_single_delete', kwargs={
                'tag_id': self._test_tag.pk
            }
        )

    def _request_test_tag_multiple_delete_view(self):
        return self.post(
            viewname='tags:tag_multiple_delete', data={
                'id_list': self._test_tag.pk
            }
        )

    def _request_test_tag_edit_view(self):
        return self.post(
            viewname='tags:tag_edit', kwargs={
                'tag_id': self._test_tag.pk
            }, data={
                'color': TEST_TAG_COLOR_EDITED,
                'label': TEST_TAG_LABEL_EDITED
            }
        )

    def _request_test_tag_list_view(self):
        return self.get(viewname='tags:tag_list')


class TaggedDocumentUploadWizardStepViewTestMixin(
    TagTestMixin, WebFormSourceTestMixin, SourceActionViewTestMixin
):
    def _request_test_source_document_upload_post_view_with_tags(self):
        tag_pk_list = list(
            Tag.objects.values_list('pk', flat=True)
        )

        return self._request_test_source_document_upload_post_view(
            extra_data={'tags': tag_pk_list}
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='Appearance'), name='appearance'
)

event_theme_created = namespace.add_event_type(
    label=_(message='Theme created'), name='theme_created'
)
event_theme_edited = namespace.add_event_type(
    label=_(message='Theme edited'), name='theme_edited'
)
event_user_theme_settings_edited = namespace.add_event_type(
    label=_(message='User theme edited'), name='user_theme_edited'
)



import logging

from rest_framework import status
from rest_framework.generics import get_object_or_404
from rest_framework.response import Response

from mayan.apps.converter.api_view_mixins import APIImageViewMixin
from mayan.apps.rest_api import generics

from ..models.trashed_document_models import TrashedDocument
from ..permissions import (
    permission_document_version_view, permission_document_view,
    permission_trashed_document_delete, permission_trashed_document_restore
)
from ..serializers.trashed_document_serializers import TrashedDocumentSerializer
from ..tasks import task_trashed_document_delete

logger = logging.getLogger(name=__name__)


class APITrashedDocumentDetailView(generics.RetrieveDestroyAPIView):
    """
    Returns the selected trashed document details.
    delete: Delete the trashed document.
    get: Retrieve the details of the trashed document.
    """
    lookup_url_kwarg = 'document_id'
    mayan_object_permission_map = {
        'DELETE': permission_trashed_document_delete,
        'GET': permission_document_view
    }
    serializer_class = TrashedDocumentSerializer
    source_queryset = TrashedDocument.objects.all()

    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()

        task_trashed_document_delete.apply_async(
            kwargs={
                'trashed_document_id': instance.pk,
                'user_id': self.request.user.pk
            }
        )

        return Response(status=status.HTTP_202_ACCEPTED)


class APITrashedDocumentListView(generics.ListAPIView):
    """
    get: Returns a list of all the trashed documents.
    """
    mayan_object_permission_map = {'GET': permission_document_view}
    ordering_fields = ('id', 'label')
    serializer_class = TrashedDocumentSerializer
    source_queryset = TrashedDocument.objects.all()


class APITrashedDocumentRestoreView(generics.ObjectActionAPIView):
    """
    post: Restore a trashed document.
    """
    lookup_url_kwarg = 'document_id'
    mayan_object_permission_map = {
        'POST': permission_trashed_document_restore
    }
    source_queryset = TrashedDocument.objects.all()

    def object_action(self, obj, request, serializer):
        obj.restore(user=self.request.user)


class APITrashedDocumentImageView(
    APIImageViewMixin, generics.RetrieveAPIView
):
    """
    get: Returns an image representation of the selected trashed document.
    """
    lookup_url_kwarg = 'document_id'
    mayan_object_permission_map = {'GET': permission_document_version_view}

    def get_object(self):
        obj = super().get_object()

        # Return a 404 if the document doesn't have any pages.
        first_page = obj.pages.first()

        if first_page:
            first_page_id = first_page.pk
        else:
            first_page_id = None

        return get_object_or_404(queryset=obj.pages, pk=first_page_id)

    def get_source_queryset(self):
        return TrashedDocument.objects.all()



from mayan.apps.appearance.classes import Icon

# Cluster

icon_setting_cluster_configuration_save = Icon(
    driver_name='fontawesome-dual', primary_symbol='sliders-h',
    secondary_symbol='check'
)
icon_setting_cluster_namespace_list = Icon(
    driver_name='fontawesome-dual', primary_symbol='sliders-h',
    secondary_symbol='folder'
)

# Namespace

icon_setting_namespace_detail = Icon(
    driver_name='fontawesome', symbol='sliders-h'
)

# Setting

icon_setting_edit = Icon(driver_name='fontawesome', symbol='pencil-alt')
icon_setting_revert = Icon(driver_name='fontawesome', symbol='undo')
icon_setting_setup = Icon(
    driver_name='fontawesome', symbol='sliders-h'
)



from django.contrib import admin

from .models import SignatureCapture


@admin.register(SignatureCapture)
class SignatureCaptureAdmin(admin.ModelAdmin):
    list_display = (
        'document', 'internal_name', 'text', 'user', 'date_time_created',
        'date_time_edited'
    )



import hashlib
from itertools import islice
import logging
import shutil

from django.apps import apps
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from mayan.apps.converter.classes import ConverterBase
from mayan.apps.converter.exceptions import (
    InvalidOfficeFormat, PageCountError
)
from mayan.apps.databases.classes import ModelQueryFields
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerMethodAfter
from mayan.apps.file_caching.models import CachePartitionFile
from mayan.apps.mime_types.classes import MIMETypeBackend

from ..classes import DocumentFileAction
from ..events import event_document_file_created, event_document_file_edited
from ..literals import (
    DOCUMENT_FILE_PAGE_CREATE_BATCH_SIZE,
    STORAGE_NAME_DOCUMENT_FILE_PAGE_IMAGE_CACHE
)
from ..settings import setting_hash_block_size
from ..signals import signal_post_document_file_upload

logger = logging.getLogger(name=__name__)


class DocumentFileBusinessLogicMixin:
    @staticmethod
    def hash_function():
        return hashlib.sha256()

    @classmethod
    def execute_pre_create_hooks(cls, kwargs=None):
        """
        Helper method to allow checking if it is possible to create
        a new document file.
        """
        cls._execute_hooks(
            hook_list=cls._hooks_pre_create, instance=None, kwargs=kwargs
        )

    @classmethod
    def register_post_save_hook(cls, func, order=None):
        cls._insert_hook_entry(
            hook_list=cls._post_save_hooks, func=func, order=order
        )

    @classmethod
    def register_pre_create_hook(cls, func, order=None):
        cls._insert_hook_entry(
            hook_list=cls._hooks_pre_create, func=func, order=order
        )

    @classmethod
    def register_pre_open_hook(cls, func, order=None):
        cls._insert_hook_entry(
            hook_list=cls._pre_open_hooks, func=func, order=order
        )

    @classmethod
    def register_pre_save_hook(cls, func, order=None):
        cls._insert_hook_entry(
            hook_list=cls._pre_save_hooks, func=func, order=order
        )

    @method_event(
        action_object='document',
        event_manager_class=EventManagerMethodAfter,
        event=event_document_file_created,
        target='self'
    )
    def _create(self, *args, **kwargs):
        DocumentFile = apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )

        self._event_keep_attributes = ('_event_actor',)
        user = getattr(self, '_event_actor', None)

        logger.info('Creating new file for document: %s', self.document)
        DocumentFile.execute_pre_create_hooks(
            kwargs={
                'document': self.document,
                'file_object': self.file.open(mode='rb'),
                'user': user
            }
        )

        try:
            self._event_ignore = True
            result = self._save(*args, **kwargs)

            logger.info(
                'New document file "%s" created for document: %s',
                self, self.document
            )

            self.document.file_latest = self
            self.document.is_stub = False

            if not self.document.label:
                self.document.label = str(self)

            self.document._event_ignore = True
            self.document.save(
                update_fields=('file_latest', 'is_stub', 'label')
            )
        except Exception as exception:
            logger.error(
                'Error creating new document file for document "%s"; %s',
                self.document, exception, exc_info=True
            )
            raise
        else:
            return result

    @method_event(
        action_object='document',
        event_manager_class=EventManagerMethodAfter,
        event=event_document_file_edited,
        target='self'
    )
    def _introspect(self):
        try:
            self.checksum_update(save=False)
            super().save(
                update_fields=('checksum',)
            )

            self.mimetype_update(save=False)
            super().save(
                update_fields=('encoding', 'mimetype',)
            )

            self.size_update(save=False)
            super().save(
                update_fields=('size',)
            )

            self.page_count_update(save=False)
        except Exception as exception:
            logger.error(
                'Error introspecting new document file for document '
                '"%s"; %s', self.document, exception, exc_info=True
            )
            raise
        else:
            self.upload_complete()

    @cached_property
    def cache(self):
        Cache = apps.get_model(app_label='file_caching', model_name='Cache')
        return Cache.objects.get(
            defined_storage_name=STORAGE_NAME_DOCUMENT_FILE_PAGE_IMAGE_CACHE
        )

    @cached_property
    def cache_partition(self):
        partition, created = self.cache.partitions.get_or_create(
            name='file-{}'.format(self.uuid)
        )
        return partition

    def checksum_update(self, save=True):
        """
        Open a document file's file and update the checksum field using
        the user provided checksum function.
        """
        DocumentFile = apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )

        block_size = setting_hash_block_size.value
        if block_size == 0:
            # If the setting value is 0 that means disable read limit.
            # To disable the read limit passing None won't work, we pass
            # -1 instead as per the Python documentation.
            # https://docs.python.org/2/tutorial/inputoutput.html#methods-of-file-objects
            block_size = -1

        if self.exists():
            hash_object = DocumentFile.hash_function()
            with self.open(raw=True) as file_object:
                while (True):
                    data = file_object.read(block_size)
                    if not data:
                        break

                    hash_object.update(data)

            self.checksum = str(
                hash_object.hexdigest()
            )

            if save:
                self.save(
                    update_fields=('checksum',)
                )

            return self.checksum

    def get_document_file_latest(self):
        return self.document.files.exclude(pk=self.pk).order_by('timestamp').only('id').last()

    def execute_pre_save_hooks(self):
        """
        Helper method to allow checking if new files are possible from
        outside the model. Currently used by the document file upload link
        condition.
        """
        DocumentFile = apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )

        DocumentFile._execute_hooks(
            hook_list=DocumentFile._pre_save_hooks, instance=self
        )

    def exists(self):
        """
        Returns a boolean value that indicates if the document's file
        exists in storage. Returns True if the document's file is verified to
        be in the document storage. This is a diagnostic flag to help users
        detect if the storage has desynchronized (ie: Amazon's S3).
        """
        name = self.file.name
        self.file.close()

        return self.file.storage.exists(name=name)

    def get_api_image_url(
        self, maximum_layer_order=None, transformation_instance_list=None,
        user=None
    ):
        first_page = self.pages.first()
        if first_page:
            return first_page.get_api_image_url(
                maximum_layer_order=maximum_layer_order,
                transformation_instance_list=transformation_instance_list,
                user=user
            )

    def get_cache_partitions(self):
        result = [self.cache_partition]
        for page in self.file_pages.all():
            result.append(page.cache_partition)

        return result

    def get_intermediate_file(self):
        cache_filename = 'intermediate_file'

        try:
            cache_file = self.cache_partition.get_file(
                filename=cache_filename
            )
        except CachePartitionFile.DoesNotExist:
            logger.debug(msg='Intermediate file not found.')

            try:
                with self.open() as file_object:
                    converter = ConverterBase.get_converter_class()(
                        file_object=file_object
                    )
                    with converter.to_pdf() as pdf_file_object:
                        with self.cache_partition.create_file(filename=cache_filename) as file_object:
                            shutil.copyfileobj(
                                fsrc=pdf_file_object, fdst=file_object
                            )

                        return self.cache_partition.get_file(filename=cache_filename).open()
            except InvalidOfficeFormat:
                return self.open()
            except Exception as exception:
                logger.error(
                    'Error creating intermediate file "%s"; %s.',
                    cache_filename, exception, exc_info=True
                )
                try:
                    cache_file = self.cache_partition.get_file(
                        filename=cache_filename
                    )
                except CachePartitionFile.DoesNotExist:
                    """Non fatal, ignore."""
                else:
                    cache_file.delete()
                raise exception
        else:
            logger.debug(msg='Intermediate file found.')
            return cache_file.open()

    def get_label(self):
        return self.filename
    get_label.short_description = _(message='Label')

    @property
    def is_in_trash(self):
        return self.document.is_in_trash

    def mimetype_update(self, save=True):
        """
        Read a document version's file and determine the mime type by using
        the MIME type backend.
        """
        if self.exists():
            try:
                with self.open() as file_object:
                    mimetype_backend = MIMETypeBackend.get_backend_instance()
                    self.mimetype, self.encoding = mimetype_backend.get_mime_type(
                        file_object=file_object
                    )
            except Exception:
                self.mimetype = ''
                self.encoding = ''
            finally:
                if save:
                    self.save(
                        update_fields=('encoding', 'mimetype')
                    )

    def open(self, raw=False):
        """
        Return a file descriptor to a document file's file irrespective of
        the storage backend.
        """
        DocumentFile = apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )

        name = self.file.name
        self.file.close()
        if raw:
            return self.file.storage.open(name=name)
        else:
            file_object = self.file.storage.open(name=name)

            result = DocumentFile._execute_hooks(
                hook_list=DocumentFile._pre_open_hooks,
                instance=self, file_object=file_object
            )

            if result:
                return result['file_object']
            else:
                return file_object

    def page_count_update(self, save=True, user=None):
        try:
            with self.open() as file_object:
                converter = ConverterBase.get_converter_class()(
                    file_object=file_object, mime_type=self.mimetype
                )
                detected_pages = converter.get_page_count()
        except PageCountError:
            """Converter backend doesn't understand the format."""
        else:
            DocumentFilePage = apps.get_model(
                app_label='documents', model_name='DocumentFilePage'
            )

            for page in self.pages.all():
                page._event_actor = user
                page._event_ignore = True
                page.delete()

            document_file_pages = (
                DocumentFilePage(
                    document_file=self, page_number=page_number + 1
                ) for page_number in range(detected_pages)
            )

            while True:
                batch = list(
                    islice(
                        document_file_pages,
                        DOCUMENT_FILE_PAGE_CREATE_BATCH_SIZE
                    )
                )

                if not batch:
                    break

                DocumentFilePage.objects.bulk_create(
                    batch_size=DOCUMENT_FILE_PAGE_CREATE_BATCH_SIZE,
                    objs=batch
                )

            if save:
                self._event_actor = user
                self.save()

            return detected_pages

    @property
    def pages(self):
        DocumentFilePage = apps.get_model(
            app_label='documents', model_name='DocumentFilePage'
        )
        queryset = ModelQueryFields.get(model=DocumentFilePage).get_queryset()
        return queryset.filter(
            pk__in=self.file_pages.values('pk')
        )

    @property
    def pages_first(self):
        return self.pages.first()

    def save_to_file(self, file_object):
        """
        Save a copy of the document from the document storage backend
        to the local filesystem.
        """
        with self.open() as input_file_object:
            shutil.copyfileobj(fsrc=input_file_object, fdst=file_object)

    def size_update(self, save=True):
        """
        Get a document file's disk file size from the storage layer and store
        it into the model.
        """
        if self.exists():
            name = self.file.name
            self.file.close()
            self.size = self.file.storage.size(name=name)

            if save:
                self.save(
                    update_fields=('size',)
                )

    def upload_complete(self):
        DocumentFile = apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )

        signal_post_document_file_upload.send(
            sender=DocumentFile, instance=self
        )

    @property
    def uuid(self):
        # Make cache UUID a mix of document UUID, file ID.
        return '{}-{}'.format(self.document.uuid, self.pk)

    def versions_new(self, action_name, comment=None, user=None):
        DocumentFileAction.get(name=action_name).execute(
            comment=comment, document=self.document, document_file=self,
            user=user
        )

    versions_new.help_text = _(
        'Controls what happens when a new document file is uploaded.'
    )



from django.apps import apps
from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.acls.permissions import (
    permission_acl_edit, permission_acl_view
)
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.classes import ModelCopy
from mayan.apps.common.menus import (
    menu_object, menu_return, menu_secondary, menu_setup
)
from mayan.apps.events.classes import EventModelRegistry, ModelEventType
from mayan.apps.navigation.classes import SourceColumn
from mayan.apps.views.column_widgets import TwoStateWidget

from .classes import QuotaBackend
from .events import event_quota_created, event_quota_edited
from .links import (
    link_quota_create, link_quota_delete, link_quota_edit, link_quota_list,
    link_quota_setup
)
from .permissions import (
    permission_quota_delete, permission_quota_edit, permission_quota_view
)


class QuotasApp(MayanAppConfig):
    app_namespace = 'quotas'
    app_url = 'quotas'
    has_rest_api = False
    has_tests = True
    name = 'mayan.apps.quotas'
    verbose_name = _(message='Quotas')

    def ready(self, *args, **kwargs):
        super().ready(*args, **kwargs)

        Group = apps.get_model(app_label='auth', model_name='Group')
        DocumentType = apps.get_model(
            app_label='documents', model_name='DocumentType'
        )
        Quota = self.get_model(model_name='Quota')
        User = get_user_model()

        EventModelRegistry.register(model=Quota)

        QuotaBackend.load_modules()

        ModelCopy(
            model=Quota, bind_link=True, register_permission=True
        ).add_fields(
            field_names=(
                'backend_path', 'backend_data', 'enabled',
            ),
        )

        ModelEventType.register(
            event_types=(event_quota_created, event_quota_edited),
            model=Quota
        )

        ModelPermission.register(
            model=Group, permissions=(permission_quota_edit,)
        )
        ModelPermission.register(
            model=DocumentType, permissions=(permission_quota_edit,)
        )

        ModelPermission.register(
            model=Quota, permissions=(
                permission_acl_edit, permission_acl_view,
                permission_quota_delete, permission_quota_edit,
                permission_quota_view
            )
        )

        ModelPermission.register(
            model=User, permissions=(permission_quota_edit,)
        )

        SourceColumn(
            attribute='backend_label', include_label=True, is_identifier=True,
            source=Quota
        )
        SourceColumn(
            attribute='backend_filters', include_label=True, source=Quota
        )
        SourceColumn(
            attribute='backend_usage', include_label=True, source=Quota
        )
        SourceColumn(
            attribute='enabled', include_label=True, is_sortable=True,
            source=Quota, widget=TwoStateWidget
        )

        menu_object.bind_links(
            links=(
                link_quota_edit, link_quota_delete,
            ), sources=(Quota,)
        )
        menu_return.bind_links(
            links=(link_quota_list,), sources=(
                Quota, 'quotas:quota_backend_selection',
                'quotas:quota_create', 'quotas:quota_list'
            )
        )
        menu_secondary.bind_links(
            links=(link_quota_create,), sources=(
                Quota, 'quotas:quota_backend_selection',
                'quotas:quota_create', 'quotas:quota_list'
            )
        )
        menu_setup.bind_links(
            links=(link_quota_setup,)
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import (
    DEFAULT_FILE_METADATA_AUTO_PROCESS,
    DEFAULT_FILE_METADATA_DRIVERS_ARGUMENTS
)
from .setting_migrations import FileMetadataSettingMigration

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='File metadata'),
    migration_class=FileMetadataSettingMigration, name='file_metadata',
    version='0002'
)
setting_auto_process = setting_namespace.do_setting_add(
    choices=('false', 'true'), default=DEFAULT_FILE_METADATA_AUTO_PROCESS,
    global_name='FILE_METADATA_AUTO_PROCESS', help_text=_(
        'Set new document types to perform file metadata processing '
        'automatically by default.'
    )
)
setting_drivers_arguments = setting_namespace.do_setting_add(
    default=DEFAULT_FILE_METADATA_DRIVERS_ARGUMENTS,
    global_name='FILE_METADATA_DRIVERS_ARGUMENTS', help_text=_(
        'Arguments to pass to the drivers.'
    )
)



import logging

from django.contrib.auth import get_user_model
from django.db.models.signals import post_save
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.menus import menu_secondary, menu_user
from mayan.apps.events.classes import ModelEventType

from .events import event_otp_disabled, event_otp_enabled
from .handlers import handler_initialize_new_user_otp_data
from .links import link_otp_detail, link_otp_disable, link_otp_enable

logger = logging.getLogger(name=__name__)


class AuthenticationOTPApp(MayanAppConfig):
    app_namespace = 'authentication_otp'
    app_url = 'authentication_otp'
    has_tests = True
    name = 'mayan.apps.authentication_otp'
    verbose_name = _(message='Authentication OTP')

    def ready(self):
        super().ready()

        User = get_user_model()

        ModelEventType.register(
            model=User, event_types=(
                event_otp_disabled, event_otp_enabled
            )
        )

        menu_secondary.bind_links(
            links=(link_otp_disable, link_otp_enable,), sources=(User,)
        )

        menu_user.bind_links(
            links=(
                link_otp_detail,
            )
        )

        post_save.connect(
            dispatch_uid='authentication_otp_handler_initialize_new_user_otp_data',
            receiver=handler_initialize_new_user_otp_data,
            sender=User
        )



import shlex

from django.apps import apps

from mayan.apps.templating.classes import Template

from .classes import MetadataLookup


class DocumentMetadataBusinessLogicMixin:
    @property
    def is_required(self):
        """
        Return a boolean value of True of this metadata instance's parent
        type is required for the stored document type.
        """
        return self.metadata_type.get_required_for(
            document_type=self.document.document_type
        )


class MetadataTypeBusinessLogicMixin:
    @staticmethod
    def comma_splitter(string):
        splitter = shlex.shlex(string, posix=True)
        splitter.whitespace = ','
        splitter.whitespace_split = True
        splitter.commenters = ''
        return [
            str(e) for e in splitter
        ]

    def get_default_value(self):
        template = Template(template_string=self.default)
        return template.render()

    def get_lookup_values(self):
        MetadataType = apps.get_model(
            app_label='metadata', model_name='MetadataType'
        )

        template = Template(template_string=self.lookup)
        return MetadataType.comma_splitter(
            template.render(
                context=MetadataLookup.get_as_context()
            )
        )

    def get_required_for(self, document_type):
        """
        Determine if the metadata type is required for the
        specified document type.
        """
        return document_type.metadata.filter(
            required=True, metadata_type=self
        ).exists()



from django.template import RequestContext
from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.views.document_views import DocumentListView
from mayan.apps.views.generics import SingleObjectListView
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from ..icons import (
    icon_workflow_runtime_proxy_document_list,
    icon_workflow_runtime_proxy_list,
    icon_workflow_runtime_proxy_state_document_list,
    icon_workflow_runtime_proxy_state_list, icon_workflow_template_list
)
from ..links import (
    link_workflow_template_create, link_workflow_template_state_create
)
from ..models import WorkflowRuntimeProxy, WorkflowStateRuntimeProxy
from ..permissions import permission_workflow_template_view


class WorkflowRuntimeProxyDocumentListView(
    ExternalObjectViewMixin, DocumentListView
):
    external_object_class = WorkflowRuntimeProxy
    external_object_permission = permission_workflow_template_view
    external_object_pk_url_kwarg = 'workflow_runtime_proxy_id'
    view_icon = icon_workflow_runtime_proxy_document_list

    def get_document_queryset(self):
        return self.external_object.get_documents()

    def get_extra_context(self):
        context = super().get_extra_context()
        context.update(
            {
                'no_results_text': _(
                    'Associate a workflow with some document types and '
                    'documents of those types will be listed in this view.'
                ),
                'no_results_title': _(
                    'There are no documents executing this workflow'
                ),
                'object': self.external_object,
                'title': _(
                    'Documents with the workflow: %s'
                ) % self.external_object
            }
        )
        return context


class WorkflowRuntimeProxyListView(SingleObjectListView):
    model = WorkflowRuntimeProxy
    object_permission = permission_workflow_template_view
    view_icon = icon_workflow_runtime_proxy_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_workflow_template_list,
            'no_results_main_link': link_workflow_template_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Create some workflows and associated them with a document '
                'type. Active workflows will be shown here and the documents '
                'for which they are executing.'
            ),
            'no_results_title': _(message='There are no workflows'),
            'title': _(message='Workflows')
        }


class WorkflowRuntimeProxyStateDocumentListView(
    ExternalObjectViewMixin, DocumentListView
):
    external_object_class = WorkflowStateRuntimeProxy
    external_object_permission = permission_workflow_template_view
    external_object_pk_url_kwarg = 'workflow_runtime_proxy_state_id'
    view_icon = icon_workflow_runtime_proxy_state_document_list

    def get_document_queryset(self):
        return self.external_object.get_documents()

    def get_extra_context(self):
        context = super().get_extra_context()
        context.update(
            {
                'object': self.external_object,
                'navigation_object_list': ('object', 'workflow'),
                'no_results_title': _(
                    'There are no documents in this workflow state'
                ),
                'title': _(
                    'Documents in the workflow "%s", state "%s"'
                ) % (
                    self.external_object.workflow, self.external_object
                ),
                'workflow': WorkflowRuntimeProxy.objects.get(
                    pk=self.external_object.workflow.pk
                )
            }
        )
        return context


class WorkflowRuntimeProxyStateListView(
    ExternalObjectViewMixin, SingleObjectListView
):
    external_object_class = WorkflowRuntimeProxy
    external_object_permission = permission_workflow_template_view
    external_object_pk_url_kwarg = 'workflow_runtime_proxy_id'
    view_icon = icon_workflow_runtime_proxy_state_list

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_main_link': link_workflow_template_state_create.resolve(
                context=RequestContext(
                    dict_={'object': self.external_object},
                    request=self.request
                )
            ),
            'no_results_text': _(
                'Create states and link them using transitions.'
            ),
            'no_results_title': _(
                'This workflow doesn\'t have any state'
            ),
            'object': self.external_object,
            'title': _(message='States of workflow: %s') % self.external_object
        }

    def get_source_queryset(self):
        return WorkflowStateRuntimeProxy.objects.filter(
            workflow=self.external_object
        )



from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin

from ..models import SignatureCapture

from .literals import (
    TEST_SIGNATURE_CAPTURE_DATA, TEST_SIGNATURE_CAPTURE_INTERNAL_NAME,
    TEST_SIGNATURE_CAPTURE_SVG, TEST_SIGNATURE_CAPTURE_TEXT,
    TEST_SIGNATURE_CAPTURE_TEXT_EDITED
)


class SignatureCaptureTestMixin(DocumentTestMixin):
    _test_object_model = SignatureCapture
    _test_object_name = '_test_signature_capture'
    auto_create_test_signature_capture = False

    def setUp(self):
        super().setUp()
        self._test_signature_captures = []
        if self.auto_create_test_signature_capture:
            self._create_test_signature_capture()

    def _create_test_signature_capture(self):
        total_test_signature_captures = len(self._test_signature_captures)
        text = '{}_{}'.format(
            TEST_SIGNATURE_CAPTURE_TEXT, total_test_signature_captures
        )

        self._test_signature_capture = SignatureCapture.objects.create(
            document=self._test_document, data=TEST_SIGNATURE_CAPTURE_DATA,
            internal_name=TEST_SIGNATURE_CAPTURE_INTERNAL_NAME,
            text=text, svg=TEST_SIGNATURE_CAPTURE_SVG,
            user=self._test_case_user
        )

        self._test_signature_captures.append(self._test_signature_capture)


class SignatureCaptureAPIViewTestMixin(SignatureCaptureTestMixin):
    def _request_test_signature_capture_create_api_view(self):
        self._test_object_track()

        response = self.post(
            viewname='rest_api:signature_capture-list', kwargs={
                'document_id': self._test_document.pk
            }, data={
                'data': TEST_SIGNATURE_CAPTURE_DATA,
                'internal_name': TEST_SIGNATURE_CAPTURE_INTERNAL_NAME,
                'svg': TEST_SIGNATURE_CAPTURE_SVG,
                'text': TEST_SIGNATURE_CAPTURE_TEXT
            }
        )

        self._test_object_set()

        return response

    def _request_test_signature_capture_delete_api_view(self):
        return self.delete(
            viewname='rest_api:signature_capture-detail',
            kwargs={
                'document_id': self._test_document.pk,
                'signature_capture_id': self._test_signature_capture.pk
            }
        )

    def _request_test_signature_capture_detail_api_view(self):
        return self.get(
            viewname='rest_api:signature_capture-detail',
            kwargs={
                'document_id': self._test_document.pk,
                'signature_capture_id': self._test_signature_capture.pk
            }
        )

    def _request_test_signature_capture_edit_api_view(
        self, extra_data=None, verb='patch'
    ):
        data = {
            'internal_name': self._test_signature_capture.internal_name,
            'text': TEST_SIGNATURE_CAPTURE_TEXT_EDITED
        }

        if extra_data:
            data.update(extra_data)

        verb = getattr(self, verb)
        return verb(
            viewname='rest_api:signature_capture-detail', kwargs={
                'document_id': self._test_document.pk,
                'signature_capture_id': self._test_signature_capture.pk
            }, data=data
        )

    def _request_test_signature_capture_list_api_view(self):
        return self.get(
            viewname='rest_api:signature_capture-list', kwargs={
                'document_id': self._test_document.pk
            }
        )


class SignatureCaptureViewTestMixin(SignatureCaptureTestMixin):
    def _request_test_signature_capture_create_view(self):
        self._test_object_track()

        response = self.post(
            viewname='signature_captures:signature_capture_create', kwargs={
                'document_id': self._test_document.pk
            }, data={
                'data': TEST_SIGNATURE_CAPTURE_DATA,
                'internal_name': TEST_SIGNATURE_CAPTURE_INTERNAL_NAME,
                'svg': TEST_SIGNATURE_CAPTURE_SVG,
                'text': TEST_SIGNATURE_CAPTURE_TEXT
            }
        )

        self._test_object_set()

        return response

    def _request_test_signature_capture_delete_view(self):
        return self.post(
            viewname='signature_captures:signature_capture_delete',
            kwargs={
                'signature_capture_id': self._test_signature_capture.pk
            }
        )

    def _request_test_signature_capture_detail_view(self):
        return self.get(
            viewname='signature_captures:signature_capture_detail',
            kwargs={
                'signature_capture_id': self._test_signature_capture.pk
            }
        )

    def _request_test_signature_capture_edit_view(self):
        return self.post(
            viewname='signature_captures:signature_capture_edit', kwargs={
                'signature_capture_id': self._test_signature_capture.pk
            }, data={
                'data': TEST_SIGNATURE_CAPTURE_DATA,
                'internal_name': TEST_SIGNATURE_CAPTURE_INTERNAL_NAME,
                'svg': TEST_SIGNATURE_CAPTURE_SVG,
                'text': TEST_SIGNATURE_CAPTURE_TEXT_EDITED
            }
        )

    def _request_test_signature_capture_list_view(self):
        return self.get(
            viewname='signature_captures:signature_capture_list', kwargs={
                'document_id': self._test_document.pk
            }
        )



import base64
import logging
from urllib.parse import quote_plus, unquote_plus

from furl import furl

from django.core.files.base import ContentFile
from django.template.defaultfilters import filesizeformat
from django.utils.encoding import force_str
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from rest_framework.reverse import reverse as rest_framework_reverse

from mayan.apps.common.menus import menu_object
from mayan.apps.converter.classes import AppImageErrorImage, ConverterBase
from mayan.apps.converter.exceptions import (
    AppImageError, InvalidOfficeFormat
)
from mayan.apps.converter.transformations import TransformationResize
from mayan.apps.navigation.classes import SourceColumn
from mayan.apps.sources.literals import STORAGE_NAME_SOURCE_CACHE_FOLDER
from mayan.apps.storage.classes import DefinedStorage

from .column_widgets import StoredFileThumbnailWidget
from .links import link_storage_file_delete, link_source_file_select

logger = logging.getLogger(name=__name__)


class SourceStoredFile:
    _initialized = False
    DEFAULT_PREVIEW_MAX_SIZE = 100000000
    IMAGE_ERROR_STAGING_FILE_TOO_LARGE = 'staging_file_too_large'

    @classmethod
    def initialize(cls):
        if not cls._initialized:
            AppImageErrorImage(
                name=SourceStoredFile.IMAGE_ERROR_STAGING_FILE_TOO_LARGE,
                template_name='source_stored_files/errors/staging_file_too_large.html'
            )

            SourceColumn(
                attribute='get_size_display',
                label=_(message='Size'), source=SourceStoredFile,
            )

            SourceColumn(
                label=_(message='Thumbnail'), source=SourceStoredFile,
                widget=StoredFileThumbnailWidget,
                html_extra_classes='text-center'
            )

            menu_object.bind_links(
                links=(link_source_file_select, link_storage_file_delete,),
                sources=(SourceStoredFile,)
            )

            cls._initialized = True

    def __init__(self, source, encoded_filename=None, filename=None):
        self.source = source
        if encoded_filename:
            self.encoded_filename = str(encoded_filename)

            try:
                self.filename = base64.urlsafe_b64decode(
                    s=unquote_plus(string=self.encoded_filename)
                ).decode('utf8')
            except UnicodeDecodeError:
                raise ValueError(
                    'Incorrect `encoded_filename` value.'
                )
        else:
            if not filename:
                raise KeyError(
                    'Supply either `encoded_filename` or `filename` when '
                    'instantiating a staging source file.'
                )
            self.filename = filename
            self.encoded_filename = quote_plus(
                base64.urlsafe_b64encode(
                    s=filename.encode('utf8')
                )
            )

    def __str__(self):
        return force_str(s=self.filename)

    @property
    def cache_filename(self):
        return '{}-{}'.format(
            self.source.model_instance_id, self.encoded_filename
        )

    def delete(self):
        # Don't include kwargs in .delete() as some backends might not
        # support them.
        try:
            self.image_cache_storage.delete(self.cache_filename)
        except FileNotFoundError:
            """No preview was yet generated."""

        self.storage_backend_instance.delete(
            name=self.get_full_path()
        )

    def generate_image(self, transformation_instance_list=None):
        # Check is transformed image is available.
        logger.debug(
            'transformations cache filename: %s', self.cache_filename
        )

        if self.image_cache_storage.exists(self.cache_filename):
            logger.debug(
                'staging file cache file "%s" found', self.cache_filename
            )
        else:
            logger.debug(
                'staging file cache file "%s" not found', self.cache_filename
            )
            image = self.get_image(
                transformation_instance_list=transformation_instance_list
            )

            # Since open "wb+" doesn't create files, check if the file
            # exists, if not then create it.
            self.image_cache_storage.save(
                content=ContentFile(content=b''), name=self.cache_filename
            )

            with self.image_cache_storage.open(name=self.cache_filename, mode='wb+') as file_object:
                file_object.write(
                    image.getvalue()
                )

        return self.cache_filename

    def get_api_image_url(
        self, maximum_layer_order=None, transformation_instance_list=None,
        request=None, user=None
    ):
        if self.get_size() > self.get_source_preview_max_size():
            raise AppImageError(
                error_name=SourceStoredFile.IMAGE_ERROR_STAGING_FILE_TOO_LARGE
            )

        final_url = furl()
        final_url.args = {'encoded_filename': self.encoded_filename}
        final_url.path = rest_framework_reverse(
            'rest_api:source_action-execute', kwargs={
                'action_name': 'file_image',
                'source_id': self.source.model_instance_id
            }, request=request
        )

        return final_url.tostr()

    def get_combined_transformation_list(
        self, maximum_layer_order=None, transformation_instance_list=None,
        user=None
    ):
        """
        Return a list of transformation containing the server side
        transformations for this object as well as transformations
        created from the arguments as transient interactive transformation.
        """
        result = [
            TransformationResize(
                height=self.source.kwargs.get('preview_height'),
                width=self.source.kwargs['preview_width']
            )
        ]

        # Interactive transformations second.
        result.extend(
            transformation_instance_list or []
        )

        return result

    def get_full_path(self):
        return self.filename

    def get_image(self, transformation_instance_list=None):
        if self.get_size() > self.get_source_preview_max_size():
            raise AppImageError(
                error_name=SourceStoredFile.IMAGE_ERROR_STAGING_FILE_TOO_LARGE
            )

        try:
            with self.open(mode='rb') as file_object:
                converter = ConverterBase.get_converter_class()(
                    file_object=file_object
                )

                try:
                    with converter.to_pdf() as pdf_file_object:
                        image_converter = ConverterBase.get_converter_class()(
                            file_object=pdf_file_object
                        )
                        page_image = image_converter.get_page()
                except InvalidOfficeFormat:
                    page_image = converter.get_page()
        except Exception as exception:
            # Cleanup in case of error.
            logger.error(
                'Error getting staging file image for file "%s"; %s',
                self.get_full_path(), exception
            )
            raise
        else:
            return page_image

    def get_size(self):
        return self.storage_backend_instance.size(
            name=self.get_full_path()
        )

    def get_size_display(self):
        return filesizeformat(
            bytes_=self.get_size()
        )

    def get_source_preview_max_size(self):
        return self.source.kwargs.get(
            'preview_max_size', SourceStoredFile.DEFAULT_PREVIEW_MAX_SIZE
        )

    @cached_property
    def image_cache_storage(self):
        return DefinedStorage.get(
            name=STORAGE_NAME_SOURCE_CACHE_FOLDER
        ).get_storage_instance()

    def open(self, mode=None):
        return self.storage_backend_instance.open(
            name=self.get_full_path(), mode=mode
        )

    @cached_property
    def storage_backend_instance(self):
        return self.source.get_storage_backend_instance()



from django.utils.module_loading import import_string

from ..literals import DEFAULT_TEST_SEARCH_BACKEND
from ..search_backends import SearchBackend
from ..settings import setting_backend_arguments


class DummySearchBackend(SearchBackend):
    feature_reindex = True


class TestSearchBackendProxy:
    _test_backend_path = None
    _test_class = None

    def __getattr__(self, attr):
        return getattr(self._backend, attr)

    def __init__(self, *args, **kwargs):
        if self._test_class:
            backend_path = getattr(
                self._test_class, '_test_search_backend_path',
                DEFAULT_TEST_SEARCH_BACKEND
            )
        else:
            backend_path = DEFAULT_TEST_SEARCH_BACKEND
            SearchBackend._disable()

        if self.__class__._test_backend_path != backend_path:
            self.__class__._test_backend_path = backend_path

            self.__class__._backend_kwargs = setting_backend_arguments.value.copy()
            self.__class__._backend_class = import_string(
                dotted_path=backend_path
            )

        self.__class__._backend_kwargs['_test_mode'] = True
        self.__class__._backend = self._backend_class(
            **self.__class__._backend_kwargs
        )

    def _search(self, *args, **kwargs):
        _skip_refresh = kwargs.pop('_skip_refresh', False)

        if not _skip_refresh:
            self._backend.refresh()

        return self._backend._search(*args, **kwargs)



from django.utils.translation import gettext_lazy as _

COMMAND_NAME_DEPENDENCIES_CHECK_VERSION = 'dependencies_check_version'
COMMAND_NAME_DEPENDENCIES_INSTALL = 'dependencies_install'
COMMAND_NAME_DEPENDENCIES_SHOW_VERSION = 'dependencies_show_version'

MAYAN_PYPI_NAME = 'mayan-edms'

MESSAGE_GREATER_THAN_SERVER = _(
    'Your version (%(version_local)s), is more recent than the published '
    'version (%(version_server)s).'
)
MESSAGE_NOT_LATEST = _(
    'The version you are using (%(version_local)s) is '
    'outdated. The latest version is %(version_server)s.'
)
MESSAGE_UNKNOWN_VERSION = _(
    'It is not possible to determine the latest version available.'
)
MESSAGE_UNEXPECTED_ERROR = _(
    'Unexpected error trying to determine the latest version available. '
    'Make sure your installation has a connection to the internet; '
    '%(exception)s'
)
MESSAGE_UP_TO_DATE = 'Your version (%(version_local)s), is up-to-date.'

PYPI_URL = 'https://pypi.org'
